/* Code to get discid for a cddb query.

  *** Linux Version ***

  $Id: discid.c,v 1.3 2001/04/14 17:54:53 jzawodn Exp $

  Copyright (c) 1998-2000 Jeremy D. Zawodny <Jeremy@Zawodny.com>

  This software is covered by the GPL.

  Is is based on code found in:

    To: code-review@azure.humbug.org.au 
    Subject: CDDB database reader 
    From: Byron Ellacott <rodent@route-qn.uqnga.org.au> 
    Date: Fri, 5 Jun 1998 17:32:40 +1000 

*/

/* Stripped net code, 'cause I only care about the discid */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <sys/types.h>

#include <linux/cdrom.h>

#include "sha1.h"

#include <id3/tag.h>

struct toc {
    int min, sec, frame, flags, lba;
} cdtoc[100];

struct trackdata {
    int min, sec;
} trackdata[100];

int first, last;

int read_toc(bool usedevice, char * device) {
    int drive = open(device, O_RDONLY | O_NONBLOCK);
    struct cdrom_tochdr tochdr;
    struct cdrom_tocentry tocentry;
    int i;

    ioctl(drive, CDROMREADTOCHDR, &tochdr);
    first = tochdr.cdth_trk0;
    last = tochdr.cdth_trk1;
    for (i = tochdr.cdth_trk0; i <= tochdr.cdth_trk1; i++) {
        tocentry.cdte_track = i;
        tocentry.cdte_format = CDROM_MSF;
        ioctl(drive, CDROMREADTOCENTRY, &tocentry);
        cdtoc[i-1].min = tocentry.cdte_addr.msf.minute;
        cdtoc[i-1].sec = tocentry.cdte_addr.msf.second - 2;
	if (cdtoc[i-1].sec < 0) {
	    cdtoc[i-1].sec += 60;
	    cdtoc[i-1].min--;
	}
	cdtoc[i-1].frame = tocentry.cdte_addr.msf.frame;
        cdtoc[i-1].frame += cdtoc[i-1].min*60*75;
        cdtoc[i-1].frame += cdtoc[i-1].sec*75;
	cdtoc[i-1].flags = ((tocentry.cdte_adr & 0x0F) << 4) |
			   (tocentry.cdte_ctrl & 0x0F);
	tocentry.cdte_format = CDROM_LBA;
	ioctl(drive, CDROMREADTOCENTRY, &tocentry);
	cdtoc[i-1].lba = tocentry.cdte_addr.lba;
    }
    tocentry.cdte_track = 0xAA;
    tocentry.cdte_format = CDROM_MSF;
    ioctl(drive, CDROMREADTOCENTRY, &tocentry);
    cdtoc[tochdr.cdth_trk1].min = tocentry.cdte_addr.msf.minute;
    cdtoc[tochdr.cdth_trk1].sec = tocentry.cdte_addr.msf.second;
    cdtoc[tochdr.cdth_trk1].frame = tocentry.cdte_addr.msf.frame;
    cdtoc[tochdr.cdth_trk1].frame += cdtoc[tochdr.cdth_trk1].min*60*75;
    cdtoc[tochdr.cdth_trk1].frame += cdtoc[tochdr.cdth_trk1].sec*75;
    cdtoc[tochdr.cdth_trk1].flags = ((tocentry.cdte_adr & 0x0F) << 4) |
				    (tocentry.cdte_ctrl & 0x0F);
    tocentry.cdte_format = CDROM_LBA;
    ioctl(drive, CDROMREADTOCENTRY, &tocentry);
    cdtoc[tochdr.cdth_trk1].lba = tocentry.cdte_addr.lba;
    close(drive);
    return tochdr.cdth_trk1;
}

int read_toc(char * mp3file)
{
    ID3_Tag id3Tag;
    ID3_Frame * id3Frame;
    ID3_Field * id3Field;
    unsigned char * tocdata;
    int i;
    
    id3Tag.Link(mp3file);
    id3Frame = id3Tag.Find(ID3FID_CDID);
    if(id3Frame == NULL) return -1;
    id3Field = id3Frame->GetField(ID3FN_DATA);
    tocdata = (unsigned char *)id3Field->GetRawBinary();

    first = tocdata[2];
    last = tocdata[3];
    for(i = first; i <= last; i++) {
	cdtoc[i-1].flags = tocdata[4+(i-1)*8+1];
	cdtoc[i-1].frame = (tocdata[4+(i-1)*8+4]<<24) | (tocdata[4+(i-1)*8+5]<<16) | (tocdata[4+(i-1)*8+6]<<8) | (tocdata[4+(i-1)*8+7]);
	cdtoc[i-1].min = cdtoc[i-1].frame/75/60;
	cdtoc[i-1].sec = cdtoc[i-1].frame/75%60;
	//cdtoc[i-1].lba = ((cdtoc[i-1].min*60)+cdtoc[i-1].sec)*75+cdtoc[i-1].frame-150;
	cdtoc[i-1].lba = cdtoc[i-1].frame;
    }
    cdtoc[last].flags = tocdata[4+last*8+1];
    cdtoc[last].frame = ((tocdata[4+last*8+4]<<24) | (tocdata[4+last*8+5]<<16) | (tocdata[4+last*8+6]<<8) | (tocdata[4+last*8+7])) + 150;
    cdtoc[last].min = cdtoc[last].frame/75/60;
    cdtoc[last].sec = cdtoc[last].frame/75%60;
    //cdtoc[last].lba = ((cdtoc[last].min*60)+cdtoc[last].sec)*75+cdtoc[last].frame-150;
    cdtoc[last].lba = cdtoc[last].frame - 150;

    return last;
}

unsigned int cddb_sum(int n) {
    unsigned int ret;

    ret = 0;
    while (n > 0) {
      ret += (n % 10);
      n /= 10;
    }
    return ret;
}

unsigned long cddb_discid(int tot_trks) {
    unsigned int i, t = 0, n = 0;

    i = 0;
    while (i < tot_trks) {
      n = n + cddb_sum((cdtoc[i].min * 60) + (cdtoc[i].sec + 2));
      i++;
    }
    t = ((cdtoc[tot_trks].min * 60) + cdtoc[tot_trks].sec) -
      ((cdtoc[0].min * 60) + cdtoc[0].sec + 2);
    return ((n % 0xff) << 24 | t << 8 | tot_trks);
}

// The following code from libmusicbrainz

unsigned char *rfc822_binary (void *src,unsigned long srcl)
{
  unsigned char *ret,*d;
  unsigned char *s = (unsigned char *) src;
  char *v = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._";
  unsigned long i = ((srcl + 2) / 3) * 4;
  d = ret = (unsigned char *) malloc ((size_t) ++i);
  for (i = 0; srcl; s += 3) {	/* process tuplets */
    *d++ = v[s[0] >> 2];	/* byte 1: high 6 bits (1) */
				/* byte 2: low 2 bits (1), high 4 bits (2) */
    *d++ = v[((s[0] << 4) + (--srcl ? (s[1] >> 4) : 0)) & 0x3f];
				/* byte 3: low 4 bits (2), high 2 bits (3) */
    *d++ = srcl ? v[((s[1] << 2) + (--srcl ? (s[2] >> 6) : 0)) & 0x3f] : '-';
				/* byte 4: low 6 bits (3) */
    *d++ = srcl ? v[s[2] & 0x3f] : '-';
    if (srcl) srcl--;		/* count third character if processed */
    if ((++i) == 15) {		/* output 60 characters? */
      i = 0;			/* restart line break count, insert CRLF */
      *d++ = '\015'; *d++ = '\012';
    }
  }
  *d = '\0';			/* tie off string */

  return ret;			/* return the resulting string */
}

unsigned char * musicbrainz_discid(int tracks) {
    SHA_INFO sha;
    int i;
    unsigned char brainzdiscid[28], temp[4];

    sha_init(&sha);
    sprintf((char *)temp, "%02X", first);
    sha_update(&sha, temp, strlen((char *)temp));
    sprintf((char *)temp, "%02X", last);
    sha_update(&sha, temp, strlen((char *)temp));
    sprintf((char *)temp, "%08lX", cdtoc[tracks].lba + 150);
    sha_update(&sha, temp, strlen((char *)temp));
    for (i = 0; i < 100; i++)
    {
	if (i != tracks)
	{
	    sprintf((char *)temp, "%08lX", cdtoc[i].lba + 150);
	    if (i > tracks) sprintf((char *)temp, "%08lX", 0);
	    sha_update(&sha, (unsigned char *)temp, strlen((char *)temp));
	}
    }
    sha_final(brainzdiscid, &sha);

    return rfc822_binary(brainzdiscid, 20);
}

char *musicbrainz_toc(int tracks)
{
    int i;
    char *toc, *temp;
    
    toc = (char *)calloc(1, 1024);
    snprintf(toc, 1024, "%d+%d+%d", first, last, cdtoc[tracks].frame);
    for(i = 0; i < tracks; i++)
    {
	temp = (char *)calloc(1, strlen(toc) + 1);
	strcpy(temp, toc);
        snprintf(toc, 1024, "%s+%d", temp, cdtoc[i].frame);
    }
    
    return toc;
}

void read_trackdata(int tracks)
{
    int i;
    
    for(i = 0; i < tracks; i++) {
        trackdata[i].min = cdtoc[i+1].min - cdtoc[i].min;
	trackdata[i].sec = cdtoc[i+1].sec - cdtoc[i].sec;
	if(trackdata[i].sec < 0) {
	    trackdata[i].sec += 60;
	    trackdata[i].min--;
	}
    }
    trackdata[i-1].sec -= 2;
    if(trackdata[i-1].sec < 0) {
	trackdata[i-1].sec += 60;
	trackdata[i-1].min--;
    }
    trackdata[i].min = cdtoc[i].min;
    trackdata[i].sec = cdtoc[i].sec;
}

int main(int argc, char* argv[]) {
    unsigned long cddbdiscid;
    int tracks, i, size, offset, number, iOpt, argCounter = 0;
    unsigned char * brainzdiscid, * tocdata;
    char * mp3file, * device = "/dev/cdrom", *toc;
    bool setid3 = false, useid3 = false, original = false, usedevice = false;
    ID3_Tag id3Tag;
    ID3_Frame * id3Frame;
    ID3_Field * id3Field;

    while (true)
    {
	int option_index = 0, iLongOpt = 0;
	static struct option long_options[] =
	{
	    { "help", no_argument, &iLongOpt, '?' },
	    { "setid3", required_argument, &iLongOpt, 's' },
	    { "useid3", required_argument, &iLongOpt, 'u' },
	    { "original", no_argument, &iLongOpt, 'o' }
	};
	iOpt = getopt_long(argc, argv, "?os:u:D:", long_options, &option_index);
	if (iOpt == -1) break;
	argCounter++;
	if(iOpt == 0) iOpt = iLongOpt;

	switch(iOpt) {
	    case '?':
		printf("usage: discid [OPTIONS]\nOutputs an organized list of information for any Audio CD.\n\n  -D [device]\tuse an alternate CD-ROM device (default: /dev/cdrom)\n  -o, --original\tdisplay in original CDDB output format (No MusicBrainz)\n  -s, --setid3 [MP3 FILE...]\tput the Table of Contents in the MCDI ID3v2 tag\n  -u, --useid3 [MP3 FILE]\tuse the Table of Contents in the MCDI ID3v2 tag\n  -?, --help\tdisplay this help and exit\n");
		return 0;
	    case 'o':
		original = true;
		break;
	    case 's':
		number = argc - argCounter - 1;
		setid3 = true;
		break;
	    case 'u':
		mp3file = optarg;
	        argCounter++;
		useid3 = true;
		break;
	    case 'D':
		device = optarg;
	        argCounter++;
		usedevice = true;
		break;
	}
    }
    
    /*for (i = 1; i < argc; i++)
    {
	if(strcmp(argv[i], "--help") == 0)
	{
            printf("usage: discid [-useid3 [MP3 FILE]] [-setid3 [MP3 FILE]] [--help]\nOutputs an organized list of information for any Audio CD.\n\n  -setid3 [MP3 FILE]\tplace the Table of Contents in the MCDI ID3v2 tag\n  -useid3 [MP3 FILE]\tuse the Table of Contents in the MCDI ID3v2 tag\n  --help\tdisplay this help and exit\n");
	    return 0;
	}
	else if(strcmp(argv[i], "-setid3") == 0 && i+1 < argc)
	{
	    number = argc-i-1;
	    setid3 = true;
	}
	else if(strcmp(argv[i], "-useid3") == 0 && i+1 < argc)
	{
	    mp3file = argv[i+1];
	    useid3 = true;
	}
    }*/
    
    useid3 ? tracks = read_toc(mp3file) : tracks = read_toc(usedevice, device);
    if(tracks == -1) {
	printf("The Table of Contents could not be read.\n");
	return 1;
    }
    cddbdiscid = cddb_discid(tracks);
    if(original)
    {
    printf ("%08lx ", cddbdiscid);
    printf ("%d ", tracks);
    for (i = 0; i < tracks; i++)
    {
	printf("%d ", cdtoc[i].frame + 150);
    }
    printf ("%d\n", cdtoc[tracks].frame/75);
    }
    else if(!original)
    {
    brainzdiscid = musicbrainz_discid(tracks);
    read_trackdata(tracks);
    printf("Table of Contents: total tracks: %d (total time %02d:%02d)\n", tracks, trackdata[tracks].min, trackdata[tracks].sec);
    printf(" header:                              {%02x %02x %02x %02x}\n", (((tracks+1)*8+2)>>8), (((tracks+1)*8+2)), first, last);
    for (i = 0; i < tracks; i++)
    {
	printf(" %02d.        %6d  (%02d:%02d)  [%02d:%02d]  {%02x %02x %02x %02x %02x %02x %02x %02x}\n", i + 1, cdtoc[i].frame, cdtoc[i].min, cdtoc[i].sec, trackdata[i].min, trackdata[i].sec, 0, cdtoc[i].flags, i + 1, 0, cdtoc[i].frame>>24, cdtoc[i].frame>>16&0xFF, cdtoc[i].frame>>8&0xFF, cdtoc[i].frame&0xFF);
    }
    printf(" lead-out:  %6d                    {%02x %02x %02x %02x %02x %02x %02x %02x}\n\n", cdtoc[tracks].frame - 150, 0, cdtoc[tracks].flags, 170, 0, (cdtoc[tracks].frame-150)>>24, (cdtoc[tracks].frame-150)>>16&0xFF, (cdtoc[tracks].frame-150)>>8&0xFF, (cdtoc[tracks].frame-150)&0xFF);
    printf("CDDB discid: 0x%08lx\n", cddbdiscid);
    printf("MusicBrainz discid: %s\n", brainzdiscid);
    toc = musicbrainz_toc(tracks);
    printf("MusicBrainz submit URL: http://mm.musicbrainz.org/bare/cdlookup.html?id=%s&tracks=%d&toc=%s\n", brainzdiscid, tracks, toc);
    free(toc);
    }
    if(setid3 == true)
    {
	while(number > 0)
	{
	mp3file = argv[argc-number];
	id3Tag.Link(mp3file);
	id3Frame = id3Tag.Find(ID3FID_CDID);
	if(id3Frame != NULL) id3Tag.RemoveFrame(id3Frame);
	id3Frame = new ID3_Frame(ID3FID_CDID);
	id3Field = id3Frame->GetField(ID3FN_DATA);
	
	size = 4 + (tracks + 1) * 8;
	tocdata = (unsigned char *)malloc(size);
	tocdata[0] = ((tracks+1)*8+2)>>8;
	tocdata[1] = (tracks+1)*8+2;
	tocdata[2] = first;
	tocdata[3] = last;
	for(i = 0;i <= tracks;i++)
	{
	    tocdata[4+i*8] = 0;
	    tocdata[4+i*8+1] = cdtoc[i].flags;
	    (i == tracks) ? tocdata[4+i*8+2] = 170 : tocdata[4+i*8+2] = i + 1;
	    tocdata[4+i*8+3] = 0;
	    (i == tracks) ? offset = 150 : offset = 0;
	    tocdata[4+i*8+4] = (cdtoc[i].frame-offset)>>24;
	    tocdata[4+i*8+5] = (cdtoc[i].frame-offset)>>16&0xFF;
	    tocdata[4+i*8+6] = (cdtoc[i].frame-offset)>>8&0xFF;
	    tocdata[4+i*8+7] = cdtoc[i].frame-offset&0xFF;
	}
	id3Field->Set(tocdata, size);
	id3Tag.AttachFrame(id3Frame);
	id3Tag.Update();
	number--;
	id3Tag = ID3_Tag();
	}
    }
    return 0;
}
